/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class org_lisp_ecl_EmbeddedCommonLisp */

#ifndef _Included_org_lisp_ecl_EmbeddedCommonLisp
#define _Included_org_lisp_ecl_EmbeddedCommonLisp
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     org_lisp_ecl_EmbeddedCommonLisp
 * Method:    start
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_org_lisp_ecl_EmbeddedCommonLisp_start
(JNIEnv *, jobject, jstring);

/*
 * Class:     org_lisp_ecl_EmbeddedCommonLisp
 * Method:    exec
 * Signature: (Ljava/lang/String;)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_org_lisp_ecl_EmbeddedCommonLisp_exec
(JNIEnv *, jobject, jstring);

/* /\* */
/*  * Class:     org_lisp_ecl_EmbeddedCommonLisp */
/*  * Method:    botExec */
/*  * Signature: (Ljava/lang/String;)Ljava/lang/String; */
/*  *\/ */
/* JNIEXPORT jstring JNICALL Java_org_lisp_ecl_EmbeddedCommonLisp_exec_ */
/*   (JNIEnv *, jobject, jstring); */

#ifdef __cplusplus
}
#endif
#endif
