/* colors */
@media (prefers-color-scheme: dark) {
    /* dark theme */
    html { color: seashell;
           background: #1A1A1A; }
    body { background: #1A1A1A; }
    th { border-bottom: 2px solid lightgray; }
    h1, h2, h3, h4, h5 { background-image: linear-gradient(to left, #202020, #3A3A3A); }
    code, var, code a { color: darkorange;
                        background: #2A2A2A; }
    a { color: seashell; }
    pre { background: #2A2A2A;
          color: seashell;
          /* mark longer code block with stripe on the left */
          border-left: 5px solid darkorange;
          padding-left: 10px; }
    pre.screen { background: #2A2A2A;
                 border: 1px solid lightgray; }
    pre.programlisting { background: #2A2A2A;
                         border-left: 1px solid lightgray;
                         border-top: 1px solid lightgray; }
	 /* we need a light background in order for the images to be readable */
	 img { background: white }
}
@media (prefers-color-scheme: light) {
    /* light theme */
    html { background: white }
    body { background: white }
    th { border-bottom: 2px solid gray; }
    h1, h2, h3, h4, h5 { background: lightgray; }
    code, var, code a { color: darkred;
                        background: whitesmoke; }
    a { color: #000; }
    pre { background: whitesmoke;
          color: black;
          /* mark longer code block with stripe on the left */
          border-left: 5px solid darkred;
          padding-left: 10px; }
    pre.screen { background: #EEE;
                 border: 1px solid black; }
    pre.programlisting { background: #EEEEEE;
                         border-left: 1px solid black;
                         border-top: 1px solid black; }
}

body {
    margin: 1em 125px 0 10%;
    line-height: 1.5em;
    padding: 0 2em 1em 2em;
    font: 13px Verdana,Arial, sans-serif
}
ul, dd, dl, dt { margin-top: 0; margin-bottom: 0; }
p, code, td, dl, dt {
    line-height: 1.5em;
}
table {
    font: inherit;
    border-collapse: collapse;
}
th, td {
    vertical-align: top;
}
h1, h2, h3 { padding-left: 15px; }
h4, h5 { padding-left: 5px; }
code, pre {
    font-size: 1em;
    font-family: monospace;
}
var {
    font-size: 1em;
}
/* links inside code appear the same as the code itself */
code a {
    font-weight: normal;
    text-decoration: none;
}
/* but get an underline when hovering */
code a:hover {
    text-decoration: underline;
}
/* ordinary links appear in bold */
a { font-weight: bold; }
pre.verbatim {
    margin: 0 0 0 0;
}
pre {
    overflow: auto;
}
pre.screen {
    font-weight: bold;
    padding: 0.5em;
}
pre.programlisting {
    padding: 0.5em;
}
div p { padding: 0 2em }
li p { padding: 0; margin: 0 }
hr { display: none; }
div.funcsynopsis p  {
    text-indent: -2em;
}
div.variablelist {
    padding: 0 2em;
}
.type, .funcsynopsis, .symbol {
    font-family: monospace;
}
.type, .symbol, .replaceable {
    white-space: nowrap;
}
