/*
 * Unicode character database.
 *
 * auto-generated, do not edit! (see contrib/unicode/)
 */
enum {
ECL_UCD_GENERAL_CATEGORY_Lu,
ECL_UCD_GENERAL_CATEGORY_Ll,
ECL_UCD_GENERAL_CATEGORY_Lt,
ECL_UCD_GENERAL_CATEGORY_Lm,
ECL_UCD_GENERAL_CATEGORY_Lo,
ECL_UCD_GENERAL_CATEGORY_Cc,
ECL_UCD_GENERAL_CATEGORY_Cf,
ECL_UCD_GENERAL_CATEGORY_Co,
ECL_UCD_GENERAL_CATEGORY_Cs,
ECL_UCD_GENERAL_CATEGORY_Mc,
ECL_UCD_GENERAL_CATEGORY_Me,
ECL_UCD_GENERAL_CATEGORY_Mn,
ECL_UCD_GENERAL_CATEGORY_Nd,
ECL_UCD_GENERAL_CATEGORY_Nl,
ECL_UCD_GENERAL_CATEGORY_No,
ECL_UCD_GENERAL_CATEGORY_Pc,
ECL_UCD_GENERAL_CATEGORY_Pd,
ECL_UCD_GENERAL_CATEGORY_Pe,
ECL_UCD_GENERAL_CATEGORY_Pf,
ECL_UCD_GENERAL_CATEGORY_Pi,
ECL_UCD_GENERAL_CATEGORY_Po,
ECL_UCD_GENERAL_CATEGORY_Ps,
ECL_UCD_GENERAL_CATEGORY_Sc,
ECL_UCD_GENERAL_CATEGORY_Sk,
ECL_UCD_GENERAL_CATEGORY_Sm,
ECL_UCD_GENERAL_CATEGORY_So,
ECL_UCD_GENERAL_CATEGORY_Zl,
ECL_UCD_GENERAL_CATEGORY_Zp,
ECL_UCD_GENERAL_CATEGORY_Zs,
};

enum {
ECL_UCD_BIDIRECTIONAL_CLASS_AL,
ECL_UCD_BIDIRECTIONAL_CLASS_AN,
ECL_UCD_BIDIRECTIONAL_CLASS_B,
ECL_UCD_BIDIRECTIONAL_CLASS_BN,
ECL_UCD_BIDIRECTIONAL_CLASS_CS,
ECL_UCD_BIDIRECTIONAL_CLASS_EN,
ECL_UCD_BIDIRECTIONAL_CLASS_ES,
ECL_UCD_BIDIRECTIONAL_CLASS_ET,
ECL_UCD_BIDIRECTIONAL_CLASS_FSI,
ECL_UCD_BIDIRECTIONAL_CLASS_L,
ECL_UCD_BIDIRECTIONAL_CLASS_LRE,
ECL_UCD_BIDIRECTIONAL_CLASS_LRI,
ECL_UCD_BIDIRECTIONAL_CLASS_LRO,
ECL_UCD_BIDIRECTIONAL_CLASS_NSM,
ECL_UCD_BIDIRECTIONAL_CLASS_ON,
ECL_UCD_BIDIRECTIONAL_CLASS_PDF,
ECL_UCD_BIDIRECTIONAL_CLASS_PDI,
ECL_UCD_BIDIRECTIONAL_CLASS_R,
ECL_UCD_BIDIRECTIONAL_CLASS_RLE,
ECL_UCD_BIDIRECTIONAL_CLASS_RLI,
ECL_UCD_BIDIRECTIONAL_CLASS_RLO,
ECL_UCD_BIDIRECTIONAL_CLASS_S,
ECL_UCD_BIDIRECTIONAL_CLASS_WS,
};
#define ECL_UCD_UPPERCASE_LIMIT 2
#define ECL_UCD_LOWERCASE_LIMIT 4
