/* AMD K7 gmp-mparam.h -- Compiler/machine parameter header file.

Copyright 1991, 1993, 1994, 2000, 2001, 2002, 2003, 2004, 2005, 2006 Free
Software Foundation, Inc.

This file is part of the GNU MP Library.

The GNU MP Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or (at your
option) any later version.

The GNU MP Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with the GNU MP Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
MA 02110-1301, USA. */

#define BITS_PER_MP_LIMB 32
#define BYTES_PER_MP_LIMB 4

/* 2083 MHz Athlon */

/* Generated by tuneup.c, 2006-03-21, gcc 4.0.2 */

#define MUL_KARATSUBA_THRESHOLD          28
#define MUL_TOOM3_THRESHOLD              98

#define SQR_BASECASE_THRESHOLD            0  /* always (native) */
#define SQR_KARATSUBA_THRESHOLD          50
#define SQR_TOOM3_THRESHOLD             113

#define MULLOW_BASECASE_THRESHOLD         4
#define MULLOW_DC_THRESHOLD              88
#define MULLOW_MUL_N_THRESHOLD          357

#define DIV_SB_PREINV_THRESHOLD           0  /* always */
#define DIV_DC_THRESHOLD                 97
#define POWM_THRESHOLD                  128

#define GCD_ACCEL_THRESHOLD               3
#define GCDEXT_THRESHOLD                 44
#define JACOBI_BASE_METHOD                1

#define USE_PREINV_DIVREM_1               1  /* native */
#define USE_PREINV_MOD_1                  1  /* native */
#define DIVREM_2_THRESHOLD                0  /* always */
#define DIVEXACT_1_THRESHOLD              0  /* always (native) */
#define MODEXACT_1_ODD_THRESHOLD          0  /* always (native) */

#define GET_STR_DC_THRESHOLD             22
#define GET_STR_PRECOMPUTE_THRESHOLD     35
#define SET_STR_THRESHOLD              3987

#define MUL_FFT_TABLE  { 592, 1184, 2432, 5632, 14336, 73728, 163840, 393216, 1572864, 6291456, 0 }
#define MUL_FFT_MODF_THRESHOLD          584
#define MUL_FFT_THRESHOLD              3840

#define SQR_FFT_TABLE  { 656, 1312, 2688, 5632, 14336, 57344, 163840, 393216, 1572864, 6291456, 0 }
#define SQR_FFT_MODF_THRESHOLD          672
#define SQR_FFT_THRESHOLD              3840
