/* ecl/configpre.h.  Generated from configure.ac by autoheader.  */

/* Define if building universal (internal helper macro) */
#undef AC_APPLE_UNIVERSAL_BUILD

/* Define to 1 if the 'closedir' function returns void instead of int. */
#undef CLOSEDIR_VOID

/* ECL_AVOID_FPE_H */
#undef ECL_AVOID_FPE_H

/* Define to 1 if we can set the stack size at runtime. */
#undef ECL_CAN_SET_STACK_SIZE

/* Allow STREAM operations to work on arbitrary objects */
#undef ECL_CLOS_STREAMS

/* Use CMU Common-Lisp's FORMAT routine */
#undef ECL_CMU_FORMAT

/* ECL_COMPLEX_FLOAT */
#undef ECL_COMPLEX_FLOAT

/* Do we use C or C++ compiler to compile ecl? */
#undef ECL_CXX_CORE

/* Do the fixed and optional arguments of a variadic function use a different
   calling convention? */
#undef ECL_C_COMPATIBLE_VARIADIC_DISPATCH

/* Stack grows downwards */
#undef ECL_DOWN_STACK

/* Use the serialization framework */
#undef ECL_EXTERNALIZABLE

/* ECL_IEEE_FP */
#undef ECL_IEEE_FP

/* ECL_LIBATOMIC_OPS_H */
#undef ECL_LIBATOMIC_OPS_H

/* ECL_LONG_LONG_BITS */
#undef ECL_LONG_LONG_BITS

/* Define if your newline is CR */
#undef ECL_NEWLINE_IS_CR

/* Define if your newline is CRLF */
#undef ECL_NEWLINE_IS_CRLF

/* ECL_SIGNED_ZERO */
#undef ECL_SIGNED_ZERO

/* ECL_SMALL_CONS */
#undef ECL_SMALL_CONS

/* ECL_SSE2 */
#undef ECL_SSE2

/* Network streams */
#undef ECL_TCP

/* Userland threads? */
#undef ECL_THREADS

/* Support for Unicode */
#undef ECL_UNICODE

/* Link in the database of Unicode names */
#undef ECL_UNICODE_NAMES

/* ECL_USE_MPROTECT */
#undef ECL_USE_MPROTECT

/* Allow loading dynamically linked code */
#undef ENABLE_DLOPEN

/* Use Boehm's garbage collector */
#undef GBC_BOEHM

/* GBC_BOEHM_GENGC */
#undef GBC_BOEHM_GENGC

/* GBC_BOEHM_PRECISE */
#undef GBC_BOEHM_PRECISE

/* Define to 1 if you have the 'alarm' function. */
#undef HAVE_ALARM

/* Define to 1 if you have the 'backtrace' function. */
#undef HAVE_BACKTRACE

/* Define to 1 if you have the 'backtrace_symbols' function. */
#undef HAVE_BACKTRACE_SYMBOLS

/* Define to 1 if you have the 'cabs' function. */
#undef HAVE_CABS

/* Define to 1 if you have the 'cabsf' function. */
#undef HAVE_CABSF

/* Define to 1 if you have the 'cabsl' function. */
#undef HAVE_CABSL

/* Define to 1 if you have the 'cacos' function. */
#undef HAVE_CACOS

/* Define to 1 if you have the 'cacosf' function. */
#undef HAVE_CACOSF

/* Define to 1 if you have the 'cacosh' function. */
#undef HAVE_CACOSH

/* Define to 1 if you have the 'cacoshf' function. */
#undef HAVE_CACOSHF

/* Define to 1 if you have the 'cacoshl' function. */
#undef HAVE_CACOSHL

/* Define to 1 if you have the 'cacosl' function. */
#undef HAVE_CACOSL

/* Define to 1 if you have the 'casin' function. */
#undef HAVE_CASIN

/* Define to 1 if you have the 'casinf' function. */
#undef HAVE_CASINF

/* Define to 1 if you have the 'casinh' function. */
#undef HAVE_CASINH

/* Define to 1 if you have the 'casinhf' function. */
#undef HAVE_CASINHF

/* Define to 1 if you have the 'casinhl' function. */
#undef HAVE_CASINHL

/* Define to 1 if you have the 'casinl' function. */
#undef HAVE_CASINL

/* Define to 1 if you have the 'catan' function. */
#undef HAVE_CATAN

/* Define to 1 if you have the 'catanf' function. */
#undef HAVE_CATANF

/* Define to 1 if you have the 'catanh' function. */
#undef HAVE_CATANH

/* Define to 1 if you have the 'catanhf' function. */
#undef HAVE_CATANHF

/* Define to 1 if you have the 'catanhl' function. */
#undef HAVE_CATANHL

/* Define to 1 if you have the 'catanl' function. */
#undef HAVE_CATANL

/* Define to 1 if you have the 'ccos' function. */
#undef HAVE_CCOS

/* Define to 1 if you have the 'ccosf' function. */
#undef HAVE_CCOSF

/* Define to 1 if you have the 'ccosh' function. */
#undef HAVE_CCOSH

/* Define to 1 if you have the 'ccoshf' function. */
#undef HAVE_CCOSHF

/* Define to 1 if you have the 'ccoshl' function. */
#undef HAVE_CCOSHL

/* Define to 1 if you have the 'ccosl' function. */
#undef HAVE_CCOSL

/* Define to 1 if you have the 'ceilf' function. */
#undef HAVE_CEILF

/* Define to 1 if you have the 'cexp' function. */
#undef HAVE_CEXP

/* Define to 1 if you have the 'cexpf' function. */
#undef HAVE_CEXPF

/* Define to 1 if you have the 'cexpl' function. */
#undef HAVE_CEXPL

/* Define to 1 if you have the 'cimag' function. */
#undef HAVE_CIMAG

/* Define to 1 if you have the 'cimagf' function. */
#undef HAVE_CIMAGF

/* Define to 1 if you have the 'cimagl' function. */
#undef HAVE_CIMAGL

/* Define to 1 if you have the 'clock_gettime' function. */
#undef HAVE_CLOCK_GETTIME

/* Define to 1 if you have the 'clog' function. */
#undef HAVE_CLOG

/* Define to 1 if you have the 'clogf' function. */
#undef HAVE_CLOGF

/* Define to 1 if you have the 'clogl' function. */
#undef HAVE_CLOGL

/* Define to 1 if you have the 'conj' function. */
#undef HAVE_CONJ

/* Define to 1 if you have the 'conjf' function. */
#undef HAVE_CONJF

/* Define to 1 if you have the 'conjl' function. */
#undef HAVE_CONJL

/* Define to 1 if you have the 'copysign' function. */
#undef HAVE_COPYSIGN

/* Define to 1 if you have the 'cosf' function. */
#undef HAVE_COSF

/* Define to 1 if you have the 'coshf' function. */
#undef HAVE_COSHF

/* Define to 1 if you have the 'cpow' function. */
#undef HAVE_CPOW

/* Define to 1 if you have the 'cpowf' function. */
#undef HAVE_CPOWF

/* Define to 1 if you have the 'cpowl' function. */
#undef HAVE_CPOWL

/* Define to 1 if you have the 'creal' function. */
#undef HAVE_CREAL

/* Define to 1 if you have the 'crealf' function. */
#undef HAVE_CREALF

/* Define to 1 if you have the 'creall' function. */
#undef HAVE_CREALL

/* Define to 1 if you have the 'csin' function. */
#undef HAVE_CSIN

/* Define to 1 if you have the 'csinf' function. */
#undef HAVE_CSINF

/* Define to 1 if you have the 'csinh' function. */
#undef HAVE_CSINH

/* Define to 1 if you have the 'csinhf' function. */
#undef HAVE_CSINHF

/* Define to 1 if you have the 'csinhl' function. */
#undef HAVE_CSINHL

/* Define to 1 if you have the 'csinl' function. */
#undef HAVE_CSINL

/* Define to 1 if you have the 'csqrt' function. */
#undef HAVE_CSQRT

/* Define to 1 if you have the 'csqrtf' function. */
#undef HAVE_CSQRTF

/* Define to 1 if you have the 'csqrtl' function. */
#undef HAVE_CSQRTL

/* Define to 1 if you have the 'ctan' function. */
#undef HAVE_CTAN

/* Define to 1 if you have the 'ctanf' function. */
#undef HAVE_CTANF

/* Define to 1 if you have the 'ctanh' function. */
#undef HAVE_CTANH

/* Define to 1 if you have the 'ctanhf' function. */
#undef HAVE_CTANHF

/* Define to 1 if you have the 'ctanhl' function. */
#undef HAVE_CTANHL

/* Define to 1 if you have the 'ctanl' function. */
#undef HAVE_CTANL

/* Define to 1 if you have the <dirent.h> header file. */
#undef HAVE_DIRENT_H

/* Define to 1 if you have the 'dladdr' function. */
#undef HAVE_DLADDR

/* Define to 1 if you have the <dlfcn.h> header file. */
#undef HAVE_DLFCN_H

/* Define to 1 if the system has the type 'double complex'. */
#undef HAVE_DOUBLE_COMPLEX

/* HAVE_ENVIRON */
#undef HAVE_ENVIRON

/* Define to 1 if you have the 'expf' function. */
#undef HAVE_EXPF

/* Define to 1 if you have the 'fabsf' function. */
#undef HAVE_FABSF

/* Define to 1 if you have the <fcntl.h> header file. */
#undef HAVE_FCNTL_H

/* feenableexcept works */
#undef HAVE_FEENABLEEXCEPT

/* Define to 1 if you have the <fenv.h> header file. */
#undef HAVE_FENV_H

/* Define to 1 if the system has the type 'float complex'. */
#undef HAVE_FLOAT_COMPLEX

/* Define to 1 if you have the <float.h> header file. */
#undef HAVE_FLOAT_H

/* Define to 1 if you have the 'floor' function. */
#undef HAVE_FLOOR

/* Define to 1 if you have the 'floorf' function. */
#undef HAVE_FLOORF

/* Define to 1 if you have the 'fork' function. */
#undef HAVE_FORK

/* Define to 1 if you have the 'frexpf' function. */
#undef HAVE_FREXPF

/* Define to 1 if you have the 'fseeko' function. */
#undef HAVE_FSEEKO

/* Define to 1 if you have the 'getcwd' function. */
#undef HAVE_GETCWD

/* Define to 1 if you have the 'gethostbyaddr' function. */
#undef HAVE_GETHOSTBYADDR

/* Define to 1 if you have the 'gethostbyname' function. */
#undef HAVE_GETHOSTBYNAME

/* Define to 1 if you have the 'getpagesize' function. */
#undef HAVE_GETPAGESIZE

/* Define to 1 if you have the 'getrusage' function. */
#undef HAVE_GETRUSAGE

/* Define to 1 if you have the 'gettimeofday' function. */
#undef HAVE_GETTIMEOFDAY

/* Define to 1 if you have the <inttypes.h> header file. */
#undef HAVE_INTTYPES_H

/* Define to 1 if you have the 'isatty' function. */
#undef HAVE_ISATTY

/* Define to 1 if you have the 'ldexpf' function. */
#undef HAVE_LDEXPF

/* HAVE_LIBFFI */
#undef HAVE_LIBFFI

/* Define to 1 if you have the 'gc' library (-lgc). */
#undef HAVE_LIBGC

/* Define to 1 if you have the 'sun' library (-lsun). */
#undef HAVE_LIBSUN

/* Define to 1 if you have the <limits.h> header file. */
#undef HAVE_LIMITS_H

/* Define to 1 if you have the <link.h> header file. */
#undef HAVE_LINK_H

/* Define to 1 if you have the 'log1p' function. */
#undef HAVE_LOG1P

/* Define to 1 if you have the 'log1pf' function. */
#undef HAVE_LOG1PF

/* Define to 1 if you have the 'log1pl' function. */
#undef HAVE_LOG1PL

/* Define to 1 if you have the 'logf' function. */
#undef HAVE_LOGF

/* Define to 1 if the system has the type 'long complex'. */
#undef HAVE_LONG_COMPLEX

/* Define to 1 if you have the 'lstat' function. */
#undef HAVE_LSTAT

/* Define to 1 if you have the <mach-o/dyld.h> header file. */
#undef HAVE_MACH_O_DYLD_H

/* Define to 1 if your system has a GNU libc compatible 'malloc' function, and
   to 0 otherwise. */
#undef HAVE_MALLOC

/* Define to 1 if you have the 'memmove' function. */
#undef HAVE_MEMMOVE

/* Define to 1 if you have the 'memset' function. */
#undef HAVE_MEMSET

/* Define to 1 if you have the 'mkdir' function. */
#undef HAVE_MKDIR

/* Define to 1 if you have the 'mkstemp' function. */
#undef HAVE_MKSTEMP

/* Define to 1 if you have a working 'mmap' system call. */
#undef HAVE_MMAP

/* Define to 1 if you have the 'nanosleep' function. */
#undef HAVE_NANOSLEEP

/* Define to 1 if you have the <ndir.h> header file, and it defines 'DIR'. */
#undef HAVE_NDIR_H

/* Define to 1 if you have the <netdb.h> header file. */
#undef HAVE_NETDB_H

/* Define to 1 if you have the <netinet/in.h> header file. */
#undef HAVE_NETINET_IN_H

/* HAVE_POSIX_RWLOCK */
#undef HAVE_POSIX_RWLOCK

/* Define to 1 if you have the 'powf' function. */
#undef HAVE_POWF

/* Define to 1 if you have the 'pthread_condattr_setclock' function. */
#undef HAVE_PTHREAD_CONDATTR_SETCLOCK

/* Define to 1 if you have the 'pthread_mutex_timedlock' function. */
#undef HAVE_PTHREAD_MUTEX_TIMEDLOCK

/* Define to 1 if the system has the type 'pthread_rwlock_t'. */
#undef HAVE_PTHREAD_RWLOCK_T

/* Define to 1 if the system has the type 'ptrdiff_t'. */
#undef HAVE_PTRDIFF_T

/* Define to 1 if you have the 'putenv' function. */
#undef HAVE_PUTENV

/* Define to 1 if you have the <pwd.h> header file. */
#undef HAVE_PWD_H

/* Define to 1 if your system has a GNU libc compatible 'realloc' function,
   and to 0 otherwise. */
#undef HAVE_REALLOC

/* Define to 1 if you have the <sched.h> header file. */
#undef HAVE_SCHED_H

/* Define to 1 if you have the 'sched_yield' function. */
#undef HAVE_SCHED_YIELD

/* Define to 1 if you have the 'select' function. */
#undef HAVE_SELECT

/* Define to 1 if you have the 'setenv' function. */
#undef HAVE_SETENV

/* Define to 1 if you have the 'sigprocmask' function. */
#undef HAVE_SIGPROCMASK

/* Define to 1 if you have the 'sinf' function. */
#undef HAVE_SINF

/* Define to 1 if you have the 'sinhf' function. */
#undef HAVE_SINHF

/* Define to 1 if you have the 'socket' function. */
#undef HAVE_SOCKET

/* Define to 1 if you have the 'sqrt' function. */
#undef HAVE_SQRT

/* Define to 1 if you have the 'sqrtf' function. */
#undef HAVE_SQRTF

/* Define to 1 if 'stat' has the bug that it succeeds when given the
   zero-length file name argument. */
#undef HAVE_STAT_EMPTY_STRING_BUG

/* Define to 1 if stdbool.h conforms to C99. */
#undef HAVE_STDBOOL_H

/* Define to 1 if you have the <stddef.h> header file. */
#undef HAVE_STDDEF_H

/* Define to 1 if you have the <stdint.h> header file. */
#undef HAVE_STDINT_H

/* Define to 1 if you have the <stdio.h> header file. */
#undef HAVE_STDIO_H

/* Define to 1 if you have the <stdlib.h> header file. */
#undef HAVE_STDLIB_H

/* Define to 1 if you have the 'strcasecmp' function. */
#undef HAVE_STRCASECMP

/* Define to 1 if you have the 'strchr' function. */
#undef HAVE_STRCHR

/* Define to 1 if you have the 'strerror' function. */
#undef HAVE_STRERROR

/* Define to 1 if you have the <strings.h> header file. */
#undef HAVE_STRINGS_H

/* Define to 1 if you have the <string.h> header file. */
#undef HAVE_STRING_H

/* Define to 1 if you have the 'strtol' function. */
#undef HAVE_STRTOL

/* Define to 1 if you have the 'system' function. */
#undef HAVE_SYSTEM

/* Define to 1 if you have the <sys/dir.h> header file, and it defines 'DIR'.
   */
#undef HAVE_SYS_DIR_H

/* Define to 1 if you have the <sys/ioctl.h> header file. */
#undef HAVE_SYS_IOCTL_H

/* HAVE_SYS_MMAN_H */
#undef HAVE_SYS_MMAN_H

/* Define to 1 if you have the <sys/ndir.h> header file, and it defines 'DIR'.
   */
#undef HAVE_SYS_NDIR_H

/* Define to 1 if you have the <sys/param.h> header file. */
#undef HAVE_SYS_PARAM_H

/* Define to 1 if you have the <sys/resource.h> header file. */
#undef HAVE_SYS_RESOURCE_H

/* Define to 1 if you have the <sys/select.h> header file. */
#undef HAVE_SYS_SELECT_H

/* Define to 1 if you have the <sys/socket.h> header file. */
#undef HAVE_SYS_SOCKET_H

/* Define to 1 if you have the <sys/stat.h> header file. */
#undef HAVE_SYS_STAT_H

/* Define to 1 if you have the <sys/time.h> header file. */
#undef HAVE_SYS_TIME_H

/* Define to 1 if you have the <sys/types.h> header file. */
#undef HAVE_SYS_TYPES_H

/* Define to 1 if you have the <sys/utsname.h> header file. */
#undef HAVE_SYS_UTSNAME_H

/* Define to 1 if you have the <sys/wait.h> header file. */
#undef HAVE_SYS_WAIT_H

/* Define to 1 if you have the 'tanf' function. */
#undef HAVE_TANF

/* Define to 1 if you have the 'tanhf' function. */
#undef HAVE_TANHF

/* Define to 1 if you have the 'times' function. */
#undef HAVE_TIMES

/* Define to 1 if you have the 'tzset' function. */
#undef HAVE_TZSET

/* Define to 1 if you have the <ulimit.h> header file. */
#undef HAVE_ULIMIT_H

/* Define to 1 if you have the 'uname' function. */
#undef HAVE_UNAME

/* Define to 1 if you have the <unistd.h> header file. */
#undef HAVE_UNISTD_H

/* Define to 1 if you have the 'vfork' function. */
#undef HAVE_VFORK

/* Define to 1 if you have the <vfork.h> header file. */
#undef HAVE_VFORK_H

/* Define to 1 if you have the <wchar.h> header file. */
#undef HAVE_WCHAR_H

/* Define to 1 if 'fork' works. */
#undef HAVE_WORKING_FORK

/* Define to 1 if 'vfork' works. */
#undef HAVE_WORKING_VFORK

/* Define to 1 if the system has the type '_Bool'. */
#undef HAVE__BOOL

/* Define to 1 if 'lstat' dereferences a symlink specified with a trailing
   slash. */
#undef LSTAT_FOLLOWS_SLASHED_SYMLINK

/* Define to the address where bug reports for this package should be sent. */
#undef PACKAGE_BUGREPORT

/* Define to the full name of this package. */
#undef PACKAGE_NAME

/* Define to the full name and version of this package. */
#undef PACKAGE_STRING

/* Define to the one symbol short name of this package. */
#undef PACKAGE_TARNAME

/* Define to the home page for this package. */
#undef PACKAGE_URL

/* Define to the version of this package. */
#undef PACKAGE_VERSION

/* Define as the return type of signal handlers ('int' or 'void'). */
#undef RETSIGTYPE

/* Define to the type of arg 1 for 'select'. */
#undef SELECT_TYPE_ARG1

/* Define to the type of args 2, 3 and 4 for 'select'. */
#undef SELECT_TYPE_ARG234

/* Define to the type of arg 5 for 'select'. */
#undef SELECT_TYPE_ARG5

/* Define to 1 if all of the C89 standard headers exist (not just the ones
   required in a freestanding environment). This macro is provided for
   backward compatibility; new code need not use it. */
#undef STDC_HEADERS

/* Define to 1 if you can safely include both <sys/time.h> and <time.h>. This
   macro is obsolete. */
#undef TIME_WITH_SYS_TIME

/* Define to 1 if your <sys/time.h> declares 'struct tm'. */
#undef TM_IN_SYS_TIME

/* Define WORDS_BIGENDIAN to 1 if your processor stores words with the most
   significant byte first (like Motorola and SPARC, unlike Intel). */
#if defined AC_APPLE_UNIVERSAL_BUILD
# if defined __BIG_ENDIAN__
#  define WORDS_BIGENDIAN 1
# endif
#else
# ifndef WORDS_BIGENDIAN
#  undef WORDS_BIGENDIAN
# endif
#endif

/* Define for Solaris 2.5.1 so the uint32_t typedef from <sys/synch.h>,
   <pthread.h>, or <semaphore.h> is not used. If the typedef were allowed, the
   #define below would cause a syntax error. */
#undef _UINT32_T

/* Define for Solaris 2.5.1 so the uint64_t typedef from <sys/synch.h>,
   <pthread.h>, or <semaphore.h> is not used. If the typedef were allowed, the
   #define below would cause a syntax error. */
#undef _UINT64_T

/* Define for Solaris 2.5.1 so the uint8_t typedef from <sys/synch.h>,
   <pthread.h>, or <semaphore.h> is not used. If the typedef were allowed, the
   #define below would cause a syntax error. */
#undef _UINT8_T

/* Define to empty if 'const' does not conform to ANSI C. */
#undef const

/* ecl_int16_t */
#undef ecl_int16_t

/* ecl_int32_t */
#undef ecl_int32_t

/* ecl_int64_t */
#undef ecl_int64_t

/* ecl_int8_t */
#undef ecl_int8_t

/* compiler understands long long */
#undef ecl_long_long_t

/* ecl_uint16_t */
#undef ecl_uint16_t

/* ecl_uint32_t */
#undef ecl_uint32_t

/* ecl_uint64_t */
#undef ecl_uint64_t

/* ecl_uint8_t */
#undef ecl_uint8_t

/* compiler understands long long */
#undef ecl_ulong_long_t

/* Define to '__inline__' or '__inline' if that's what the C compiler
   calls it, or to nothing if 'inline' is not supported under any name.  */
#ifndef __cplusplus
#undef inline
#endif

/* Define to the type of a signed integer type of width exactly 16 bits if
   such a type exists and the standard includes do not define it. */
#undef int16_t

/* Define to the type of a signed integer type of width exactly 32 bits if
   such a type exists and the standard includes do not define it. */
#undef int32_t

/* Define to the type of a signed integer type of width exactly 64 bits if
   such a type exists and the standard includes do not define it. */
#undef int64_t

/* Define to the type of a signed integer type of width exactly 8 bits if such
   a type exists and the standard includes do not define it. */
#undef int8_t

/* Define to rpl_malloc if the replacement function should be used. */
#undef malloc

/* Define as a signed integer type capable of holding a process identifier. */
#undef pid_t

/* Define to rpl_realloc if the replacement function should be used. */
#undef realloc

/* Define as 'unsigned int' if <stddef.h> doesn't define. */
#undef size_t

/* Define to the type of an unsigned integer type of width exactly 16 bits if
   such a type exists and the standard includes do not define it. */
#undef uint16_t

/* Define to the type of an unsigned integer type of width exactly 32 bits if
   such a type exists and the standard includes do not define it. */
#undef uint32_t

/* Define to the type of an unsigned integer type of width exactly 64 bits if
   such a type exists and the standard includes do not define it. */
#undef uint64_t

/* Define to the type of an unsigned integer type of width exactly 8 bits if
   such a type exists and the standard includes do not define it. */
#undef uint8_t

/* Define as 'fork' if 'vfork' does not work. */
#undef vfork

/* Define to empty if the keyword 'volatile' does not work. Warning: valid
   code using 'volatile' can become incorrect without. Disable with care. */
#undef volatile
